#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include <math.h>
#include <curses.h>
#include "fbtetrisai.h"

pos calibrationCorner;
pos gridCorner;
piece currentPiece;
piece nextPiece[5] = {_, _, _, _, _};
int field[WIDTH][HEIGHT];

// piece info
pos currentPos;
int currentRot;

// AI
int targetXPos, targetYPos;
int targetRot;

piece getGridPieceTypeFromColor(int r, int g, int b) {
	if(r > 95  && r < 140 && g > 85  && g < 115 && b > 5   && b < 50)  return O;
	if(r > 5   && r < 30  && g > 85  && g < 115 && b > 95  && b < 120) return I;
	if(r > 50  && r < 75  && g > 95  && g < 120 && b > 5   && b < 30)  return S;
	if(r > 85  && r < 115 && g > 0   && g < 35  && b > 0   && b < 35)  return Z;
	if(r > 90  && r < 135 && g > 45  && g < 85  && b > 0   && b < 35)  return L;
	if(r > 5   && r < 35  && g > 20  && g < 55  && b > 95  && b < 125) return J;
	if(r > 95  && r < 130 && g > 0   && g < 30  && b > 105 && b < 136) return T;
	
	return _;
}

piece getNextPieceTypeFromColor(int r, int g, int b) {
	if(r == 119 && g == 110 && b == 19)  return O;
	if(r == 31  && g == 96  && b == 116) return I;
	if(r == 168 && g == 250 && b == 103) return S;
	if(r == 253 && g == 101 && b == 100) return Z;
	if(r == 255 && g == 186 && b == 119) return L;
	if(r == 105 && g == 149 && b == 255) return J;
	if(r == 253 && g == 103 && b == 250) return T;
	
	return _;
}

int isActivePiece(int r, int g, int b) {
	if(r == 122 && g == 106 && b == 31) return 1;
	if(r == 122 && g == 16 && b == 127) return 1;
	if(r == 108 && g == 18 && b == 18) return 1;
	if(r == 20 && g == 51 && b == 116) return 1;
	if(r == 61 && g == 113 && b == 13) return 1;
	if(r == 16 && g == 99 && b == 117) return 1;
	if(r == 131 && g == 69 && b == 18) return 1;
	
	return 0;
}

int isValidPiece(int type) {
	return (type < 8 && type >= 0);
}

char pieceToChar(piece type) {
	assert(isValidPiece(type));
	return *("OISZLJT_"+type);
}

void updateNextPiece() {
	piece detectedPiece;
	int r, g, b;
	
	getPixelColor(calibrationCorner.x + NEXT_PIECE_OFFSET_X, calibrationCorner.y + NEXT_PIECE_OFFSET_Y, &r, &g, &b);
	detectedPiece = getNextPieceTypeFromColor(r, g, b);
	
	if(isValidPiece(detectedPiece)) {
		currentPiece = nextPiece[0];    
		nextPiece[0] = nextPiece[1];
		nextPiece[1] = nextPiece[2];
		nextPiece[2] = nextPiece[3];
		nextPiece[3] = nextPiece[4];
		nextPiece[4] = detectedPiece;
	}
}

void updateField() {
	int x, y, r, g, b;
	
	for(y=0; y<HEIGHT; y++) {
		for(x=0; x<WIDTH; x++) {
			getPixelColor(gridCorner.x + 22*x, gridCorner.y + 22*y, &r, &g, &b);
			if(!isActivePiece(r, g, b)) {
				field[x][HEIGHT-1-y] = getGridPieceTypeFromColor(r, g, b);
			}
			else {
				field[x][HEIGHT-1-y] = _;
			}
			
		}
	}
	
}

void initCurses() {
	initscr();
	keypad(stdscr, TRUE);
	nonl();
	cbreak();
	curs_set(0);
	noecho();
	nodelay(stdscr, TRUE);
}

// inputs: currentPiece, nextPiece[], field[][]
// outputs: targetXPos, targetRot
void performAILogic() {
	int possible;
	int x = 0, y = 1337, ytry, rot, cell, cell2, score, neighbour;
	int topScore = -999999, topScoreX = 0, topScoreY = 0, topScoreRot = 0;
	
	/*
	Temporary random test:
		targetXPos = rand()%WIDTH;
		targetRot = rand()%4;
	*/
	
	// For every X position
	for(x=0; x<WIDTH; x++) {
		// For every possible rotation of the current piece
		for(rot=0; rot<rotations[currentPiece]; rot++) {
			possible = 1;
			
			// Check wether block is inside a wall
			for(cell=0; cell<8; cell+=2) {
				if(x+pieces[currentPiece][rot][cell] < 0 || x+pieces[currentPiece][rot][cell] > WIDTH-1) {
					possible = 0;
					break;
				}
			}
			
			// Don't try this if it's not possible
			if(!possible) {
				continue;
			}
			
			// Try to move piece as far down as possible
			for(ytry=HEIGHT-1; ytry>=0; ytry--) {
				possible = 1;
				
				for(cell=0; cell<8; cell+=2) {
					// Are we inside the ground?
					if(ytry+pieces[currentPiece][rot][cell+1] < 0) {
						possible = 0;
					}
				
					// Only check if we're not in the ceiling
					else if(ytry+pieces[currentPiece][rot][cell+1] < HEIGHT) {
						// Are we inside other blocks?
						if(field[x+pieces[currentPiece][rot][cell]][ytry+pieces[currentPiece][rot][cell+1]] != _) {
							possible = 0;
						}
					}
					
					if(!possible) {
						break;
					}
				}
				
				if(!possible) {
					break;
				}
			}
			y = ytry+1;
			
			// Piece did not fit anywhere
			if(y >= HEIGHT) {
				continue;
			}
			
			// Lowest possible spot is now (x, y){rot}
			// Calculate amount of touching edges per cell
			score = 0;
			for(cell=0; cell<8; cell+=2) {
				// Right wall
				if(x+1+pieces[currentPiece][rot][cell] == WIDTH) {
					score+=3;
				}
				else {
					// Field piece on the right
					if(field[x+1+pieces[currentPiece][rot][cell]][y+pieces[currentPiece][rot][cell+1]] != _) {
						score+=2;
					}
				}
				
				// Left wall
				if(x-1+pieces[currentPiece][rot][cell] == -1) {
					score+=3;
				}
				else {
					// Field piece on the left
					if(field[x-1+pieces[currentPiece][rot][cell]][y+pieces[currentPiece][rot][cell+1]] != _) {
						score+=2;
					}
				}
				
				// Floor
				if(y+pieces[currentPiece][rot][cell+1] == 0) {
					score+=5;
				}
				else {
					if(field[x+pieces[currentPiece][rot][cell]][y-1+pieces[currentPiece][rot][cell+1]] != _) {
						score+=2;
						//printf("Bottom field piece on cell=%d x=%d rot=%d. (found piece %c on field at (%d,%d))\n", cell, x, rot, \
						// pieceToChar(field[x+pieces[currentPiece][rot][cell]][y-1+pieces[currentPiece][rot][cell+1]]), x+pieces[currentPiece][rot][cell], y-1+pieces[currentPiece][rot][cell+1]);
					}
					
					else {
						//printf("Cell %d on x=%d and rot=%d is floating.\n", cell, x, rot);
						neighbour = 0;
						for(cell2=0; cell2<8; cell2+=2) {
							if((pieces[currentPiece][rot][cell2+1] == pieces[currentPiece][rot][cell+1] - 1) && (pieces[currentPiece][rot][cell2] == pieces[currentPiece][rot][cell])) {
								neighbour = 1;
								
								break;
							}
						}
						
						// Punishment for empty block beneith piece
						if(!neighbour) {
							score-=50;							
						}
					}
					
				}
			}
			
			// Test to make it prefer low spots
			score -= 2*y;
			
			
			// We now have the score for the current state (x, y){rot} in 'score'
			
			if(score > topScore) {
				topScore = score;
				topScoreX = x;
				topScoreY = y;
				topScoreRot = rot;
			}
			
			//printf(" Attempt: x:%d, rot:%d, score:%d\n", x, rot, score);
		}
	}
	
	// Set the target to the state with the highest score.
	targetXPos = topScoreX;
	targetYPos = topScoreY;
	targetRot = topScoreRot;
	
	printf("piece:%c, score:%d, x:%d, y:%d, rot:%d\n", pieceToChar(currentPiece), topScore, topScoreX, topScoreY, topScoreRot);
}

// This could be done more cleanly, but this works fine I guess
int targetClearsLines() {
	int cell, x, y, filled, ans = 0;
	
	// Insert piece into field
	for(cell=0; cell<8; cell+=2) {
		field[targetXPos+pieces[currentPiece][targetRot][cell]][targetYPos+pieces[currentPiece][targetRot][cell+1]] = currentPiece;
	}
	
	// Check for filled lines
	for(y=0; y<HEIGHT; y++) {
		filled = 0;
		
		for(x=0; x<WIDTH; x++) {
			filled += (field[x][y] != _);
		}
		
		// All collumns filled, we have a line
		if(filled == WIDTH) {
			ans = 1;
			break;
		}
	}
	
	// Remove piece from field
	for(cell=0; cell<8; cell+=2) {
		field[targetXPos+pieces[currentPiece][targetRot][cell]][targetYPos+pieces[currentPiece][targetRot][cell+1]] = _;
	}
	
	return ans;
	
}

int main(int argc, char *argv[]) {
	int r, g, b;
	int i, j;
	int lastr, lastg, lastb;
	int checkedNextPiece = 0;
	piece type;

	initXStuff();

	if(argc == 3) {
		calibrationCorner.x = atoi(argv[1]);
		calibrationCorner.y = atoi(argv[2]);
	}
	else {
		// Wait for user to position their mouse
		usleep(5000000);
		getMousePosition(&calibrationCorner.x, &calibrationCorner.y);
		
		printf("Corner calibrated to: (%d, %d)\n", calibrationCorner.x, calibrationCorner.y);
		
		exit(0);
	}
	
	
	//initCurses();
	
	gridCorner.x = (calibrationCorner.x + 11);
	gridCorner.y = (calibrationCorner.y - 5);
	
	// Start with a clean field
	updateField();
	
	// Initialize piece data
	currentPos.x = 4;
	currentPos.y = HEIGHT-1;
	currentRot = 0;
	
	while(1) {
		// Update next piece info
		updateNextPiece();
		
		// If no current piece is known, determine it
		if(currentPiece == _) {
			do {
				getPixelColor(gridCorner.x + 22*4, gridCorner.y + 22*0, &r, &g, &b);
				if(!isActivePiece(r, g, b)) getPixelColor(gridCorner.x + 22*4, gridCorner.y + 22*1, &r, &g, &b);
				if(!isActivePiece(r, g, b)) getPixelColor(gridCorner.x + 22*4, gridCorner.y + 22*2, &r, &g, &b);
				if(!isActivePiece(r, g, b)) getPixelColor(gridCorner.x + 22*4, gridCorner.y + 22*3, &r, &g, &b);
				if(!isActivePiece(r, g, b)) getPixelColor(gridCorner.x + 22*4, gridCorner.y + 22*4, &r, &g, &b);
				if(!isActivePiece(r, g, b)) getPixelColor(gridCorner.x + 22*4, gridCorner.y + 22*5, &r, &g, &b);
				
			} while(!isActivePiece(r, g, b));
			currentPiece = getGridPieceTypeFromColor(r, g, b);
		}
		else {
			usleep(20000);
		}
		
		// inputs: currentPiece, nextPiece[], field[][], currentPos, currentRot
		performAILogic();
		// outputs: targetXPos, targetRot
		
		//printf("target x:%d, rot:%d\n", targetXPos, targetRot);
		
		// React accordingly
		while(currentRot != targetRot || currentPos.x != targetXPos) {
			if(currentRot < targetRot) {
				sendKeyStroke(ACTION_ROTATE_RIGHT);
				currentRot++;
				//printf("rotated right\n");
			}
			else if(currentRot > targetRot) {
				sendKeyStroke(ACTION_ROTATE_LEFT);
				currentRot--;
				//printf("rotated left\n");
			}
			
			if(currentPos.x < targetXPos) {
				currentPos.x++;
				sendKeyStroke(ACTION_MOVE_RIGHT);
			}
			else if(currentPos.x > targetXPos) {
				currentPos.x--;
				sendKeyStroke(ACTION_MOVE_LEFT);
			}
			usleep(40000);
		}
		
		// We're now positioned correctly (hopefully), so let's do a hard drop
		sendKeyStroke(ACTION_HARD_DROP);

		
		// Make sure it doesn't glitch because of the moving field
		if(targetClearsLines()) {
			// TODO: Optimize this delay (as low as possible)
			usleep(400000);
			printf("Cleared a line!\n");
		}
		
		
		
		// Reset values for next piece
		currentPos.x = 4;
		currentPos.y = HEIGHT-1;
		currentRot = 0;
		
		/*
		// Exit on any key presses
		if(getch() != ERR) {
			endwin();
			break;
		}
		
	
		// Some debug info
		mvprintw(0, 0, "Current: %c, Upcoming: %c, %c, %c, %c, %c", \
			pieceToChar(currentPiece), pieceToChar(nextPiece[0]), pieceToChar(nextPiece[1]), \
			pieceToChar(nextPiece[2]), pieceToChar(nextPiece[3]), pieceToChar(nextPiece[4]));
		*/
		
		// Update the field and print it
		updateField();
		//for(i=0; i<HEIGHT; i++) {
		//	for(j=0; j<WIDTH; j++) {
		//		mvprintw(1+i, j*2, "%c", pieceToChar(field[j][HEIGHT-1-i]));
		//	}
		//	
		//}
		
		// Redraw ncurses screen
		//refresh();
	}
	
	
	
	// Some debug loops here
	/*
	while(1) {
		getPixelColor(gridCorner.x + 22*4, gridCorner.y + 22*3, &r, &g, &b);
		if(lastr != r && lastg != g && lastb != b) {
			printf("%d, %d, %d\n", r, g, b);
		}
		lastr = r; lastg = g; lastb = b;
	}
	*/
	
	/*
	while(1) {
		sendKeyStroke(ACTION_HARD_DROP);
		usleep(200000); // TODO: find the perfect value for this number
		updateNextPiece();
		usleep(3000000);
	}*/
	
	
	freeXStuff();
	return 0;
}




